// This may look like C code, but it is really -*- C++ -*-
/*
 ************************************************************************
 *
 *			   UNT Virtual Machine
 *			 Central Processing Unit
 *
 * The present file implemements a central processing unit, i.e. emulates
 * all the operations the "regular" CPU would perform.
 * Once started, the CPU runs until it gets stopped or trap occures. In
 * the latter case, 'trap_signal' is raised so the trap handler can
 * wake up and handle the trap. The trap handler is not a part of CPU,
 * though, and runs as a separate thread within the "hardware".
 *
 ************************************************************************
 */

#pragma implementation
#include "hardware.h"
#include "myenv.h"
#include <std.h>

/*
 *------------------------------------------------------------------------
 *		   	   Initialization
 */

int _ebx_save;				// newproc() spoils %ebx on the
					// second return. So we have to
					// save it.

CentralProcessingUnit::CentralProcessingUnit(Memory& _memory)
	: MemoryManagementUnit(_memory),
	  stopped("CPU operation",0),
	  trap_signal("CPU trap",0)
{
  pc = 0;
  clock = -1;
  trap_code = NONE;

//  message("\ntrap_signal ptr %x, val =%x",&trap_signal,*(int *)&trap_signal);
  asm("movl %ebx,__ebx_save");		// Saving %ebx. It's a shame we need
					// to resort to such a kludge!
  if( !newproc("CPU process",1) )
  {					// This is a main CPU process
    for(;;)
    {
      stopped--;			// Wait until can run again
      if( execute_instruction() )
	stopped++;			// If everything was fine, keep going
      else
	trap_signal++;
    }
  }

  asm("movl __ebx_save,%ebx");		// Restore %ebx. I wish I didn't
					// have to do this!
//  message("\ntrap_signal ptr %x, val =%x",&trap_signal,*(int *)&trap_signal);
}

/*
 *------------------------------------------------------------------------
 *		      Starting and stopping the CPU
 */

void CentralProcessingUnit::start(void)
{
  if( !is_running() )
    single_message("CPU: has been stopped, starting..."),
    stopped++;
  else
    single_message("CPU: already running");
}

void CentralProcessingUnit::stop(void)
{
  if( is_running() )
    single_message("CPU: has been running, stopping..."),
    stopped--;
  else
    single_message("CPU: already stopped");
}

/*
 *------------------------------------------------------------------------
 *		   Emulates a single CPU operation
 *	   Fetch an instruction at current pc and execute it
 * Returns 0 if trap or other interrupt has occurred that makes continuation
 * useless. Otherwise returns 1.
 * Note, that pc always points to the instruction following the one
 * which has been executed or which has caused the interrupt.
 *
 */

int CentralProcessingUnit::execute_instruction(void)
{
  trap_code = NONE;

  if( clock == 0 )			// Handle the clock and generate
  {					// interrupt if necessary
    trap_code = CLOCKINT;		// (Clock is to turn negative)
    --clock;
    return 0;
  }
  else if( clock > 0 )
    --clock;

  inst_length = 0;				// Fetch and examine the
  pc++; inst_length++;				// 1st word of the instruction
  if( *(Word *)&iword1 = fetch(pc-1), got_fault() )
  {
    trap_code = MEMORY;
    pc -= inst_length;				// Set pc to repeat the 
    return 0;					// operation after recovery
  }
  opcode = (OpCodes)iword1.opcode;
  Word& reg_a = (*this)[iword1.areg];
  Word  reg_b = (*this)[iword1.breg];
  EA = 0;
  if( opcode >= TWO_WORD )
  {					// Fetch the second word of instruction
    ++pc, ++inst_length;
    if( *(Word *)&iword2 = fetch(pc-1), got_fault() )
    {
      trap_code = MEMORY;
      pc -= inst_length;			// Set pc to repeat the 
      return 0;					// operation after recovery
    }
    EA = iword2.ea;
					// Fetch the indirect address
    if( iword2.indirect && (EA = get(EA), got_fault() ) )
    {
      trap_code = MEMORY;
      pc -= inst_length;			// Set pc to repeat the 
      return 0;					// operation after recovery
    }
    if( iword1.breg != 0 )		// Add the index to the address
      EA += reg_b;
  }

				// Print the trace info
  begin_printing();
  message("\nCPU: Executing instruction");
  message("\n  PC  Clock Opcode Areg Breg  R(Areg) R(Breg) SvcOp Ind   EA"
	  "  C(EA)"
	  "\n%4ob %3d    %2d    %2d   %2d    %06o  %06o",
	  pc-inst_length,clock,opcode,iword1.areg,iword1.breg,
	  reg_a,reg_b);
  if( opcode >= TWO_WORD )
    message("    %2d   %1d %4ob %06o",iword2.svcop,iword2.indirect,EA,
	    get(EA)), clear_error();	// Clear possible error due to get()
  message("\n");
  end_printing();

				// Interpret the instruction
  switch( opcode )
  {
    case HALT:
         trap_code = ILLEGOP;
	 return 0;

    case DUMP:
	 Context::dump();
	 dump_phys_memory(reg_a,reg_b);
	 break;

    case LOADR:
	 reg_a = reg_b;
	 break;

    case ADD:
         reg_a += reg_b;
	 break;

    case SUBT:
         reg_a -= reg_b;
	 break;

    case MULT:
         reg_a *= reg_b;
	 break;

    case DIV:
	 if( reg_b != 0 )
	   reg_a /= reg_b;
	 break;

    case AND:
         reg_a &= reg_b;
	 break;

    case OR:
         reg_a |= reg_b;
	 break;

    case XOR:
         reg_a ^= reg_b;
	 break;

    case SHIFTL:
         reg_a <<= reg_b;
	 break;

    case SHIFTR:
         reg_a >>= reg_b;
	 break;

    case NOP:
	 break;

    case IOPR:
         trap_code = ILLEGOP;
	 return 0;

    case LOAD:
	 register Word ea_cont = get(EA);
	 if( got_fault() )
	 {
	   trap_code = MEMORY;
	   pc -= inst_length;			// Set pc to repeat the 
	   return 0;				// operation after recovery
	 }
	 reg_a = ea_cont;
	 break;

    case LOADI:
	 reg_a = EA;
	 break;

    case STORE:
	 if( put(EA,reg_a), got_fault() )
	 {
	   trap_code = MEMORY;
	   pc -= inst_length;			// Set pc to repeat the 
	   return 0;				// operation after recovery
	 }
	 break;

    case BR:
	 pc = EA;
	 break;

    case BRNEG:
	 if( reg_a < 0 )
	   pc = EA;
	 break;

    case BRZERO:
	 if( reg_a == 0 )
	   pc = EA;
	 break;

    case BRPOS:
	 if( reg_a > 0 )
	   pc = EA;
	 break;

    case BRSUBR:
	 reg_a = pc;
	 pc = EA;
	 break;

    case READ:
    case WRITE:
         trap_code = ILLEGOP;
	 return 0;

    case SVC:
	 svcop = iword2.svcop;
         trap_code = SVCCALL;
	 return 0;

    default:
         trap_code = ILLEGOP;
	 return 0;
  }
  return 1;
}


/*
 *------------------------------------------------------------------------
 * 			   Dump the context
 */

void CentralProcessingUnit::dump(void) const
{
  Context::dump();
  MemoryManagementUnit::dump();
}

void Context::dump(void) const
{
  register int i;
  message("\n\t\t\tContext Dump\n");
  message("\n   R0     R1     R2     R3     R4     R5     R6     R7\n");
  for(i=0; i<=7; i++)
    message(" %06o",Registers[i]);
  message("\n\n   R8     R9     R10    R11   R12    R13    R14    R15\n");
  for(i=8; i<=15; i++)
    message(" %06o",Registers[i]);
  message("\n\nPC = %06o\n\n",pc);
}

/*
 *------------------------------------------------------------------------
 *		     Processor context functions
 */

				// Clean up the context at the very beginning
Context::Context(void)
{
  memset(Registers,0,sizeof(Registers));
  pc = 0;
}

				// Get ref to a given register
Word& Context::operator [] (const int reg_number)
{
  assure( reg_number >= 0 && reg_number <= 15, "Bad register no" );
  return Registers[reg_number];
}

				// Copy the context from another one
void Context::copy_context(const Context& a_context)
{
  memcpy(this,&a_context,sizeof(Context));
}


