// This may look like C code, but it is really -*- C++ -*-
/*
 ************************************************************************
 *
 *			   UNT Virtual Machine
 *
 * 		This is the core of the operating system
 *
 * The functions defined in the present file are executed within the
 * TRAP-handler thread that gets control if CPU raised the trap signal.
 * While CPU is stopped during the trap handling, the trap thread
 * handles traps and system requests (SVC traps) and performs all the
 * high level process scheduling.
 *
 ************************************************************************
 */

#pragma implementation "cpu_manager.h"
#include "cpu_manager.h"
#include "io_manager.h"
#include "myenv.h"

/*
 *------------------------------------------------------------------------
 *			Initializing the operating system
 */

CPUManager::CPUManager(CentralProcessingUnit& _CPU,
		       MemoryManager& _mem_manager,IOManager& _io_manager)
	: CPU(_CPU),
	  memory_manager(_mem_manager),
	  io_manager(_io_manager),
	  Halt("Computer halt",0),
	  ProcessTable(20)
{
  semas = new SemaphoreTable(20);
  running_pcb = 0;
  if( !newproc("TRAP handling",1) )
  {					// This is a CPU trap handler
    for(;;)
    {
      CPU.wait_for_trap();		// Wait for the trap
      if( trap_handler(CPU.q_trap_code()) == DISPATCH )
	dispatch();
      CPU.start();
    }
  }
}


				// Load a program from the drum and run it
void CPUManager::commence(void)
{
  int no_programs = memory_manager.q_no_programs();
  console("There are %d programs on the drum to run",no_programs);
  register int i;
  for(i=0; i<no_programs; i++)
  {
    PCB& pcb = (*this)[new_pid()];
    pcb.brand_new();
    pcb.pc = memory_manager.load_program(pcb,i);
    readyPCBs.append(pcb);
  }
  dispatch();
  CPU.start();
}

/*
 *------------------------------------------------------------------------
 *			    TRAP handler
 */

HANDLER_STATUS
CPUManager::trap_handler(const CentralProcessingUnit::TRAPCODE trap_code)
{
  save_context();			// Save the CPU context anyway
  switch(trap_code)
  {
    case CentralProcessingUnit::ILLEGOP:
         if( CPU.q_opcode() == HALT )
	   console("Current process has terminated normally (by HALT)");
	 else if( CPU.q_opcode() == READ || CPU.q_opcode() == WRITE )
	   return initiate_IO(CPU.q_opcode(),CPU.q_EA());
	 else
	   console("Illegal operation (dec code %d)",CPU.q_opcode());
	 return terminate();
    
    case CentralProcessingUnit::SVCCALL:
	 return svc_handler((SVCCODE)CPU.q_svcop());

    case CentralProcessingUnit::CLOCKINT:
	 single_message("Clock Interrupt");
	 if( readyPCBs.is_empty() )
	   return RESUME;		// We've got only a single process
	 readyPCBs.append(*running_pcb);
	 return DISPATCH;		// Else pick up smth else

    case CentralProcessingUnit::MEMORY:
	 return mfault_handler(CPU.what_fault());

    default:
         _error("FATAL: Trap handler has been entered with illegal "
		"trap code %d",trap_code);
  }
  return DISPATCH;
}

/*
 *------------------------------------------------------------------------
 *			    SVC Handler
 */

HANDLER_STATUS CPUManager::svc_handler(const SVCCODE svc_code)
{
  single_message("SVC interrupt %d",svc_code);
  switch(svc_code)
  {
    case DUMP_STATUS:			// Dump the OS status
         dump();
         return RESUME;
    
    case FORK:				// Fork a process.
					// reg_a := PID of a kid or parent
	 return fork(CPU.q_RegA(),CPU.q_EA());	// EA = start address of a kid

    case WAIT_KIDS:			// Wait for kid processes to terminate
	 return wait_for_kids();

    case SEMINIT:			// Create a semaphor for a process
         return create_semaphore(CPU.q_RegA(),CPU.q_EA());

    case SEM_P:				// P-operation on semaphore
	 return semaphore_p((SID)CPU[CPU.q_RegA()]);

    case SEM_V:				// V-operation on semaphore
	 return semaphore_v((SID)CPU[CPU.q_RegA()]);

    case KILL:				// Try to kill the process
	 return shoot((PID)CPU[CPU.q_RegA()]);

    default:
         console("Illegal SVC call %d, program is being tossed out",
		 svc_code);
	 return terminate();
  }
  return DISPATCH;
}

/*
 *------------------------------------------------------------------------
 *		  Elementary process scheduling functions
 */

				// Prepare the process for running
				// on the CPU
void CPUManager::prepare_for_running(void)
{
  assert( running_pcb != 0 );
  assert( !CPU.is_running() );
  (*running_pcb).load_context(CPU);
  CPU.clock = Time_quant;
}

				// Save the status of the currently
				// running process as the process is
				// going to lose the CPU control
void CPUManager::save_context(void)
{
  assert( running_pcb != 0 );
  assert( !CPU.is_running() );
  (*running_pcb).save_context(CPU);
}

				// Terminate the currently running process
HANDLER_STATUS CPUManager::terminate(void)
{
  console("Terminating the process PID %d",running_pcb->id);
  (*running_pcb).dump();
  kill(*running_pcb);
  return DISPATCH;			// Pick up a new process to run
}

				// Pick up a new process to run and make
				// it current
void CPUManager::dispatch(void)
{
  if( q_all_free() )			// Check if all processes are trough
    Halt++;				
  running_pcb = &(*this)[readyPCBs.get_from_head()->id]; // Implies waiting
  single_message("Dispatchung process %d...",running_pcb->id);
  prepare_for_running();
}

/*
 *------------------------------------------------------------------------
 *		      Dump whatever goes on in the system
 */

void CPUManager::dump(void)
{
  begin_printing();
  message("\n%s\n\n\t\t\tOperating System Status\n",_Minuses);
  message("\nCurrent process\n");
  (*running_pcb).dump();
  ProcessTable::dump();
  (*semas).dump();
  memory_manager.dump_status();
  io_manager.dump();
  end_printing();
}

/*
 *------------------------------------------------------------------------
 *			   Making new processes
 */

				// Create a child process at EA of
				// the currently running process
				// Put PID of the son into the parent rega,
				// and PID of the parent into the son rega
HANDLER_STATUS CPUManager::fork(const int rega,Address EA)
{
  assert( running_pcb != 0 );
  if( running_pcb -> lchild != NIL_pid && running_pcb -> rchild != NIL_pid )
  {
    console("ABEND: attempt to create the 3d child");
    return terminate();
  }

  PID son_id = new_pid();
  if( son_id == NIL_pid )
  {
    console("ABEND: can't create a process - too many are running");
    return terminate();
  }

  PCB& son_pcb = (*this)[son_id];
  PCB& dad_pcb = *running_pcb;
  single_message("Creating a child %d for a parent %d",son_pcb.id,dad_pcb.id);

  son_pcb.fork_from_dad(dad_pcb);
  son_pcb[rega] = dad_pcb.id;
  son_pcb.pc    = EA;

  dad_pcb[rega] = son_pcb.id;

  if( !memory_manager.fork_son((MMContext&)son_pcb,(const MMContext&)dad_pcb) )
  {					// If some problem occurred during  
    kill(son_pcb);			// memory allocation for the son
    return RESUME;
  }
  
  readyPCBs.append(dad_pcb);
  readyPCBs.append(son_pcb);
  return DISPATCH;
}

				// Wait until all the kids of the
				// running process are through.
				// Return RESUME if the running process
				// has got no kids.
HANDLER_STATUS CPUManager::wait_for_kids(void)
{
  assert( running_pcb != 0 );
  if( running_pcb->lchild == NIL_pid && running_pcb->rchild == NIL_pid )
    return RESUME;				// No kids
  running_pcb->status = PCB::Wait_for_kids;	// The PCB remains unqueued
  return DISPATCH;
}

				// Try to kill the process
HANDLER_STATUS CPUManager::shoot(const PID pid)
{
  single_message("An attempt to shoot a process %d",pid);
  if( pid == running_pcb->id )
  {
    console("The current process %d shot himself",pid);
    return terminate();
  }

  if( pid == NIL_pid )
  {
    console("An attempt to kill a nonexistent process");
    return terminate();
  }

  if( pid != running_pcb->lchild && pid != running_pcb->rchild )
  {
    console("Process %d has no right to shoot %d",running_pcb->id,pid);
    return terminate();
  }

  PCB& victim = (*this)[pid];
  if( victim.status == PCB::Doing_io )
    victim.status = PCB::Shall_die;

  kill(victim);
  return RESUME;
}

/*
 *------------------------------------------------------------------------
 *				Kill the process
 * The program performs the clean-up and releases all the resources
 * that process occupied.
 * 	- Kids are terminated
 * 	- Parent is notified and turned ready if has been waiting
 * 	- Owned semaphores are destroyed and all the processes being 
 *	  waiting on it are terminated
 * 	- the process is purged of all semaphore waiting lists if
 *	  it has been waiting on P operation
 *	- dispose of the memory of the deseased process
 */

void CPUManager::kill(PCB& pcb)
{
  assert(pcb.status != PCB::Dead);

  readyPCBs.purge_id(pcb.id);		// Purge from the ready queue
					// (if any)
  if( pcb.lchild != NIL_pid )
    kill((*this)[pcb.lchild]);

  if( pcb.rchild != NIL_pid )
    kill((*this)[pcb.rchild]);

  if( pcb.parent != NIL_pid )
  {
    PCB& dad_pcb = (*this)[pcb.parent];
    if( dad_pcb.lchild == pcb.id )
      dad_pcb.lchild = NIL_pid;
    else if( dad_pcb.rchild == pcb.id )
      dad_pcb.rchild = NIL_pid;
    else
      _error("FATAL: parent doesn't know of his son");

				// If dad has been waiting for kids, it may go
    if( dad_pcb.lchild == NIL_pid && dad_pcb.rchild == NIL_pid &&
       dad_pcb.status == PCB::Wait_for_kids )
      dad_pcb.status = PCB::Ok,
      readyPCBs.append(dad_pcb);
    pcb.parent = NIL_pid;
  }

					// Release all owned semaphores
  SID sid;
  while( (sid = (*semas).owned_by(pcb.id)) != NIL_sid )
  {
    (*semas).dispose(sid);
  }

  if( pcb.status == PCB::Wait_on_sema )
    (*semas).purge(pcb.id);

  memory_manager.dispose((MMContext&)pcb);
  dispose(pcb.id);
}

/*
 *------------------------------------------------------------------------
 *			Processes and Semaphores
 */
				// Create a new semaphore with initial value EA
				// Put SID in rega
				// Return RESUME if everything is fine
HANDLER_STATUS CPUManager::create_semaphore(const int rega,Address EA)
{
  assert( running_pcb != 0 );

  SID semid = (*semas).new_semaphore(EA,running_pcb->id);
  if( semid == NIL_sid )
  {
    console("ABEND: Can't create a semaphore - too many are in use");
    return terminate();
  }

  single_message("Semaphore %d (in_value %d) has been allocated for PID %d",
		 semid,EA,running_pcb->id);

  CPU[rega] = semid;
  return RESUME;
}

				// Check to see that the Sema is valid
				// to perform P or V operation on.
				// It should be active and belong to
				// the running process or its ancestor
Sema * CPUManager::valid_semaphore(const SID sid)
{
  if( !(*semas).is_active(sid) )
    return 0;
  Sema& sem = (*semas)[sid];
  PID owner = sem.q_owner();			// Check ownership
  PID id    = running_pcb->id;			// through the chain of
  for(; id != NIL_pid; id = (*this)[id].parent)	// ancestorship
    if( id == owner )
      return &sem;
  return 0;
}

				// Semaphore P-operation
				// Return RESUME if the process is to
				// be resumed
HANDLER_STATUS CPUManager::semaphore_p(const SID sid)
{
  single_message("\nP-operation on semaphore %d",sid);
  assert( running_pcb != 0 );
  Sema * semp;
  if( (semp = valid_semaphore(sid)) == 0 )
  {
    console("Semaphore %d may not be used by process %d",sid,
	    running_pcb->id);
    return terminate();
  }

  if( !(*semp).p(*running_pcb) )
  {					// Suspend the current process
    single_message("Process %d should wait on semaphore %d",
		   running_pcb->id,sid);
    running_pcb->status = PCB::Wait_on_sema;
    return DISPATCH;
  }  

  return RESUME;
}


				// Semaphore V-operation
				// Return RESUME if the process is to
				// be resumed
HANDLER_STATUS CPUManager::semaphore_v(const SID sid)
{
  single_message("V-operation on semaphore %d",sid);
  assert( running_pcb != 0 );
  Sema * semp;
  if( (semp = valid_semaphore(sid)) == 0 )
  {
    console("Semaphore %d may not be used by process %d",sid,
	    running_pcb->id);
    return terminate();
  }

  PID pid_to_wake;
  if( (pid_to_wake = (*semp).v()) != NIL_pid )
  {					// Wake up pid_to_wake
    single_message("Process %d is being woken up",pid_to_wake);
    PCB& pcb_to_wake = (*this)[pid_to_wake];
    pcb_to_wake.status = PCB::Ok;
    readyPCBs.append(pcb_to_wake);
  }  

  return RESUME;
}

/*
 *------------------------------------------------------------------------
 *			Preliminary Memory Fault Handling
 * Advance memory handling is done by the MemoryManager
 */

HANDLER_STATUS CPUManager::mfault_handler
	(const MemoryManagementUnit::MemoryFault code)
{
  MMUContext::VirtualMemoryAddr culprit_VA = CPU.q_current_VA();

  switch(code)
  {
    case MemoryManagementUnit::OutofRange:
         console("ABEND: Virtual address %o is out of range",culprit_VA);
	 return terminate();

    case MemoryManagementUnit::WrongPageTable:
         console("ABEND: Page table has wrong format for VA %o",culprit_VA);
	 return terminate();

    case MemoryManagementUnit::WrongVPageNo:
         console("ABEND: Wrong virtual page number in VA %o",culprit_VA);
	 return terminate();

    case MemoryManagementUnit::PageInvalidated:
	 break;					// Try to handle that

    case MemoryManagementUnit::NoAccess:
         console("ABEND: Access to the virtual page denied, VA %o",culprit_VA);
	 return terminate();

    case MemoryManagementUnit::ReadOnly:
         console("ABEND: May not write to a read-only page, VA %o",culprit_VA);
	 return terminate();

    case MemoryManagementUnit::ExecOnly:
         console("ABEND: EXEConly page may only be fetched, VA %o",culprit_VA);
	 return terminate();

    default:
	 _error("FATAL: unknown memory fault %d, cannot handle",code);
  }

  single_message("Missing the page at VA %6o",*(Address*)&culprit_VA);

					// We've access to the page not in
					// the physical memory. Try to bring
					// it to there
  enum {Begin, Blocked, Ready} swap_status = Begin;
  for(;;)
    switch(memory_manager.load_the_page((MMContext&)*running_pcb,culprit_VA))
    {
      case MemoryManager::Ok:
           CPU.clear_error();
           (*running_pcb).load_context(CPU);
           return RESUME;

      case MemoryManager::PhysMemoryExhausted:
	   message("\nMemory exhausted: ");
	   PID victim_pid;
	   switch(swap_status)
	   {
	     case Begin:
	          start_scan();
		  swap_status = Blocked;

	     case Blocked:
		  if( (victim_pid = next_pcb(ProcessTable::Blocked)) 
		      != NIL_pid )
		  {
		    PCB& pcb = (*this)[victim_pid];
		    message("Blocked PID %d is to be swapped out",victim_pid);
		    memory_manager.swap_out((MMContext&)pcb);
		    continue;
		  }
	          start_scan();
		  swap_status = Ready;
		  
	     case Ready:
		  if( (victim_pid = next_pcb(ProcessTable::Ready))
		      != NIL_pid )
		  {
		    if( victim_pid == running_pcb->id )
		      continue;
		    PCB& pcb = (*this)[victim_pid];
		    message("Ready PID %d is to be swapped out",victim_pid);
		    memory_manager.swap_out((MMContext&)pcb);
		    continue;
		  }
		  console("We've tried hard enough, but there is no "
			  "free physical memory");
		  return terminate();
	   }
	   break;

      case MemoryManager::LethalFault:
	   return terminate();
    }
}

/*
 *------------------------------------------------------------------------
 *	The following utilities help handle the parameter block
 * the process has prepared in the memory
 * They operate in the current process virtual memory. Page table is
 * assumed to be loaded.
 * The program return FALSE if some problem occured. Usually it is the
 * reason for process termination.
 */

				// Read a word from the (virtual)
				// address space of a current process
int CPUManager::read_word(const Address va, Word& dest)
{
  if( dest = CPU.get(va), CPU.got_fault() )
    if( mfault_handler(CPU.what_fault()) != RESUME )
      return 0;				// Incorrectable memory fault

  if( dest = CPU.get(va), CPU.got_fault() )
    return 0;				// Second fault - also bad
  return 1;
}

				// Write a word to the (virtual)
				// address space of a current process
int CPUManager::write_word(const Address va, Word src)
{
  if( CPU.put(va,src), CPU.got_fault() )
    if( mfault_handler(CPU.what_fault()) != RESUME )
      return 0;				// Incorrectable memory fault

  if( CPU.put(va,src), CPU.got_fault() )
    return 0;				// Second fault - also bad
  return 1;
}
				// Lock the page of the current process
				// at the specified VA
int CPUManager::lock_page(Address va)
{
					// First, try to read this address
					// and load the page if necessary
  if( CPU.get(va), CPU.got_fault() )
    if( CPU.what_fault() != MemoryManagementUnit::PageInvalidated || 
        mfault_handler(CPU.what_fault()) != RESUME )
      return 0;				// Incorrectable memory fault
  return memory_manager.lock_loaded_page((MMContext&)*running_pcb,va);
}

				// Unlock the page of a given process
				// at the specified VA
void CPUManager::unlock_page(const PID pid,Address va)
{
  memory_manager.unlock_page((MMContext&)(*this)[pid],va);
}

				// Translate the virtual address to physical
				// one using the translation tables of the
				// running process
				// running process. The page has to be loaded
				// Returns 0 if fails
Address CPUManager::translate_va(Address va)
{
  return memory_manager.translate_va((MMContext&)*running_pcb,va);
}

				// Read/write a word from/to the 
				// PHYSICAL address space
				// Error is fatal
void CPUManager::read_word_phys_memory(const Address ra, Word& dest)
{
  Memory::MemoryAnswer ans = CPU.memory[ra];
  if( ans.out_of_range )
    _error("CPU reading phys memory failed at %o",ra);
  dest = ans.contents;
}

void CPUManager::write_word_phys_memory(const Address ra, Word src)
{
  Memory::MemoryAnswer ans = CPU.memory[ra];
  if( ans.out_of_range )
    _error("CPU writing phys memory failed at %o",ra);
  ans.contents = src;
}

/*
 *------------------------------------------------------------------------
 *			Initiate the Input/Output
 * The following program performs only preliminary I/O initiation
 */

HANDLER_STATUS CPUManager::initiate_IO(const OpCodes opcode,Address IO_params)
{
  single_message("Processing I/O parameters");
  if( !io_manager.submit_request(running_pcb->id,opcode,IO_params) )
    return terminate();
  running_pcb->status = PCB::Doing_io;
  return DISPATCH;
}

				// Post the completion of the I/O request
void CPUManager::io_request_completed(const PID pid)
{
  PCB& pcb = (*this)[pid];
  assert( pcb.status == PCB::Doing_io );
  pcb.status = PCB::Ok;
  readyPCBs.append(pcb);
}
