// This may look like C code, but it is really -*- C++ -*-
/*
 ************************************************************************
 *
 *			   UNT Virtual Machine
 *
 * 			  Input/Output managers
 *
 * A single I/O channel managers is almost an exact copy of the CPU manager.
 * It picks up the first request from its to_serve request queue, initiates
 * the channel and handles the channel interrupts. When there is nothing
 * to serve the manager waits until some request shows up in the queue.
 * The I/O manager is a separate thread. The fact that the request
 * block looks (and acts) like the Process Control Block for CPU completes
 * the analogy. There is some differences, though. The process running on the
 * CPU can be suspended (because it requested some service from OS or
 * just the CPU managers feels so) and then run again. But the channel program
 * once initiated, cannot be interrupted until it completes (or fails).
 * In either case, when the channel program is over, it is over and cannot
 * be resumed. Therefore, there is no need to save the  channel context
 * in the IORequest block.
 *
 ************************************************************************
 */

#pragma implementation "io_manager.h"
#include "io_manager.h"
#include "cpu_manager.h"
#include "myenv.h"

/*
 *------------------------------------------------------------------------
 *			Initializing the IO manager
 */

IOManager::IOManager(CPUManager& _cpu_manager)
	: requests(5),
	  cpu_manager(_cpu_manager)
{
  register int i;
  for(i=0; i<max_no_channels; i++)
    channel_managers[i] = 0;
}

void IOManager::register_channel
	(const int chan_no, IOChannelManager * _channel_manager)
{
  assert( chan_no > 0 && chan_no <= max_no_channels );
  assert( channel_managers[chan_no-1] == 0 );
  channel_managers[chan_no-1] = _channel_manager;
  message("\nRegistered channel %d",chan_no);
}

				// Tell what's going on with the manager
void IOManager::dump(void) const
{
  register int i;			// Dump status of all registered
  for(i=0; i<max_no_channels; i++)	// channels
    if( channel_managers[i] != 0 )
      channel_managers[i]->dump();
}

				// Reserve the physical page for the system
				// Returns the physical address it starts
Address IOManager::reserve_physical_page()
{
  return cpu_manager.reserve_physical_page();
}

/*
 *------------------------------------------------------------------------
 *			   Initiating the I/O
 *
 * The following functions are executed as a part of the CPU manager thread.
 * Note, that during the initiation the requesting process is still current
 *
 */

int IOManager::submit_request
	(const PID pid,const OpCodes opcode,Address IO_params)
{
  RID rid = requests.new_request();
  if( rid == NIL_rid )
  {
    console("No free request blocks to serve the process %d"
	    "\nThe process has to die",pid);
    return 0;
  }

  IORequest& req = requests[rid];

  req.pid = pid;			// Filling out the request block
  if( opcode == READ )
    req.operation = IORequest::IO_READ;
  else if( opcode == WRITE )
    req.operation = IORequest::IO_WRITE;
  else
  {
    console("Invalid I/O operation");
    return 0;
  }

  if( !cpu_manager.read_word(IO_params,req.device_no) ||
     					// Write the completion code 0
					// just to test the writability
      !cpu_manager.write_word(req.return_code_vptr=IO_params+1,0) || 
      !cpu_manager.read_word(IO_params+2,req.buffer_ptr) ||
      !cpu_manager.read_word(IO_params+3,req.rec_count) ||
      !cpu_manager.read_word(IO_params+4,req.disk_address) )
    return 0;

  message("\nInitiating I/O");
  req.dump();

  if( !cpu_manager.lock_page(req.return_code_vptr) )
    return 0;
  if( (req.return_code_pptr = cpu_manager.translate_va(req.return_code_vptr))
      == 0 )				// Try to translate the address
    return 0;				// where to write the return code to
  req.return_code = 0;

  scrutinize_request(req);
  queue_request(req);
  
  return 1;
}

				// Scrutinize the request and check its
				// validity. Lock all the pages to be taking
				// part into I/O and translate their virtual
				// addresses into the physical one.
				// In case of error, just set the return_code
				// and schedule to queue to
				// any channel handler (for termination)
				// Since the process initiating the I/O is
				// still current, we still can enjoy
				// address translation and page locking
int IOManager::scrutinize_request(IORequest& req)
{
  if( req.rec_count >= req.max_no_recs )
  {
    console("I/O request by pid #%d: too many pages, %d, are involved",
	    req.pid,req.rec_count);
    req.return_code = -1;
    return 0;
  }
  
  if( req.buffer_ptr & (Memory::pagesize-1) )
  {
    console("I/O request by pid #%d: buffer address, %o, is misaligned",
	    req.pid,req.buffer_ptr);
    req.return_code = -1;
    return 0;
  }
  
  register int i;
  register Address addr = req.buffer_ptr;
  for(i=0; i<req.rec_count; i++,addr += Memory::pagesize)
    if( !cpu_manager.lock_page(addr) || 
        (req.io_pages[i] = cpu_manager.translate_va(addr)) == 0 )
    {
      console("I/O request by pid #%d: failed to lock and translate addr %o",
	      req.pid,addr);
      req.return_code = -1;
      return 0;
    }

  return 1;
}

				// Queue the request to the appropriate
				// channel. If device_no is invalid,
				// set the return code and queue to the
				// first channel to take care of the error
				// later
void IOManager::queue_request(IORequest& req)
{
  if( req.device_no >= max_no_channels || req.device_no == 0 || 
      channel_managers[req.device_no-1] == 0 )
  {
    console("I/O request by pid #%d: device %d is invalid or inoperational",
	    req.pid,req.device_no);
    req.return_code = -1;
    req.device_no = 1;			// Make the IOChannel manager do all
  }					// the clean-up
  channel_managers[req.device_no-1]->submit_request(req);
}

/*
 *------------------------------------------------------------------------
 *		    Handle the completion of the request
 *
 * The following function is executed within the I/O Channel manager thread
 */

void IOManager::request_completed(IORequest& req)
{
  single_message("I/O request of pid %d is completed with code %d",
		 req.pid,req.return_code);
  cpu_manager.write_word_phys_memory(req.return_code_pptr,req.return_code);
  cpu_manager.unlock_page(req.pid,req.return_code_vptr);

  if( req.io_pages[0] != 0 )
  {					// Some pages were locked
    assert(req.rec_count < req.max_no_recs );
    assert( (req.buffer_ptr & (Memory::pagesize-1)) == 0 );

    register int i;
    register Address addr = req.buffer_ptr;
    for(i=0; i<req.rec_count && req.io_pages[i] != 0; 
	i++,addr += Memory::pagesize)
      cpu_manager.unlock_page(req.pid,addr),
      req.io_pages[i] = 0;
  }

  cpu_manager.io_request_completed(req.pid);
  requests.dispose(req.id);
}

/*
 *------------------------------------------------------------------------
 *		       	    A channel manager
 */

IOChannelManager::IOChannelManager(IOChannel& _channel,IOManager& _io_manager)
	: channel(_channel),
	  io_manager(_io_manager)
{
  serviced_request = 0;
  io_manager.register_channel(channel.q_channel_no(),this);
  ccw_area = io_manager.reserve_physical_page();

  if( !newproc("Channel managing",1) )
  {					// This is a channel manager thread
    for(;;)
    {
					// Wait for the request
      if( initiate_io(to_serviceIORs.get_from_head()->id) )
      {
	channel.wait_for_interrupt();		// Wait for the I/O completion
	serviced_request->return_code = channel.q_status();
      }
      io_manager.request_completed(*serviced_request);
      serviced_request = 0;
    }
  }
}

				// Tell what's going on
void IOChannelManager::dump(void) const
{
  message("\nChannel %d: ",channel.q_channel_no());

  if( serviced_request == 0 )
    message("Idling\n");
  else
    message("Processing the I/O request for pid #%d\n",serviced_request->pid),
    message("There are %d more requests in the queue\n",
	    to_serviceIORs.q_no_elems()),
    channel.dump();
}


				// Build the channel program and initiate
				// i/o for the request. 
				// Return FALSE if error was found and I/O
				// hasn't been initiated
int IOChannelManager::initiate_io(RID rid)
{
  IORequest& req = io_manager.requests[rid];
  assert( serviced_request == 0 );
  serviced_request = &req;

  message("\nChannel %d: selecting the request of pid %d for execution\n",
	  channel.q_channel_no(),req.pid);

  if( req.return_code != 0 )
    return 0;				// Request has been cancelled

					// Constructing a channel program
  CCW ccw;
  Address ccw_ptr = ccw_area;

  ccw.opcode = SEEKOP;
  ccw.keep_going = 1;
  ccw.page_no = req.disk_address;
  write_word_phys_memory(ccw_ptr++,*(Word*)&ccw);

  register int i;
  for(i=0; i<req.rec_count; i++)
  {
    ccw.opcode = (req.operation == IORequest::IO_READ ? READOP : WRITEOP);
    ccw.keep_going = (i+1 == req.rec_count ? 0 : 1);
    ccw.page_no = req.io_pages[i] / Memory::pagesize;
    write_word_phys_memory(ccw_ptr++,*(Word*)&ccw);
  }
  assert(i<Memory::pagesize);

  channel.start(ccw_area);
  return 1;
}

				// Read/write a word from/to the 
				// PHYSICAL address space
				// Error is fatal
void IOChannelManager::read_word_phys_memory(const Address ra, Word& dest)
{
  Memory::MemoryAnswer ans = channel.memory[ra];
  if( ans.out_of_range )
    _error("IOChannel %d reading phys memory failed at %o",
	   channel.q_channel_no(),ra);
  dest = ans.contents;
}

void IOChannelManager::write_word_phys_memory(const Address ra, Word src)
{
  Memory::MemoryAnswer ans = channel.memory[ra];
  if( ans.out_of_range )
    _error("IOChannel %d reading phys memory failed at %o",
	   channel.q_channel_no(),ra);
  ans.contents = src;
}
