// This may look like C code, but it is really -*- C++ -*-
/*
 ************************************************************************
 *
 *			   UNT Virtual Machine
 *				I/O Manager
 *
 * Handles I/O devices and I/O requests asynchronously of the
 * CPU manager. There is a separate I/O manager thread associated with
 * each device. Communication between the CPU manager and I/O manager
 * is asynchronous and based on queue of request blocks.
 *
 ************************************************************************
 */

#ifndef _io_manager_h
#define _io_manager_h 1
#pragma interface

#include "hardware.h"
#include "sysqueues.h"
#include "processes.h"

/*
 *------------------------------------------------------------------------
 *			    I/O requests
 * I/O request presents a unit of work for the I/O channel manager.
 * It keeps all the information necessary to initiate the I/O and reply
 * the results back.
 */

typedef unsigned short RID;		// I/O Request ID
const RID NIL_rid = 0;			// Other RIDs are positive

class IORequestTable;
class IOManager;
class IOChannelManager;

class IORequest : public QueueLink
{
  friend class IORequestTable;
  friend class IOManager;
  friend class IOChannelManager;

  PID pid;				// PID of the process requested I/O
  enum {IO_READ, IO_WRITE} operation;	// The following describes the I/O
  Word device_no;			// to perform
  Address buffer_ptr;
  Word rec_count;
  Word disk_address;
  Address return_code_vptr;		// This is a virtual memory ptr
  Address return_code_pptr;		// This is a physical memory ptr
  int return_code;			// Return code of the operation

  const int max_no_recs = 8;		// Physical addresses of the pages
  Address io_pages[max_no_recs];	// to transmit during the I/O

public:
  IORequest(void);
  ~IORequest(void) {}
  void dump(void) const;		// Tell what's going on
  void dispose(void);			// Dispose of the given request block
};


class IORequestTable
{
  IORequest * reqs;			// Array of IOrequests
  const int nslots;			// No. of request slots

  BasicQueue freeIORs;			// Queue of free IORequests

public:
  IORequestTable(const int no_slots);
  ~IORequestTable(void) {}

  IORequest& operator [] (const RID pid) const;	// Get a IORequest by its id
  RID new_request(void);		// Create a new request
  void dispose(const RID rid);		// Dispose of the request
  void dump(void) const;		// Tell what's going on
};

/*
 *------------------------------------------------------------------------
 *			Input/output request managers
 */

				// Class IOManager provides the data structures
				// and function which are used both by CPU
				// manager and I/O Channel manager threads to
				// talk to each other.
class IOChannelManager;
class CPUManager;

class IOManager : public DiagnosticPanel
{
  CPUManager& cpu_manager;

  const int max_no_channels = 4;	// Note, the channel no. starts from 1
  IOChannelManager * channel_managers[max_no_channels];

public:
  IORequestTable requests;

  IOManager(CPUManager& _cpu_manager);
  ~IOManager(void) {}
  void dump(void) const;		// Tell what's going on

					// Register a channel manager with
					// the IO manager
  void register_channel(const int chan_no,IOChannelManager * _channel_manager);

				// The following procedures are used by
				// the CPU manager thread to submit the request
					// Return FALSE in case of problem
  int submit_request(const PID pid,const OpCodes opcode,Address IO_params);
  int scrutinize_request(IORequest& req);// and perform locking and translation
  void queue_request(IORequest& req);	// and queue the request

				// The following procedures are used by the
				// I/O channel manager thread to reply back
  void request_completed(IORequest& req);// Handle the request completion

  Address reserve_physical_page();	// Reserve a page and give its address
};

				// Manager of a channel
class IOChannelManager
{
  IOChannel& channel;			// Channel to handle
  IOManager& io_manager;		// General I/O manager
  IORequest * serviced_request;		// The IO request being serviced

  BasicQueue to_serviceIORs;		// Queue of requests to service

  Address ccw_area;			// Reserved page of the physical
					// memory for constructing CCW

public:
  IOChannelManager(IOChannel& _channel, IOManager& _io_manager);
  ~IOChannelManager(void) {}

  void dump(void) const;		// Tell what's going on
  void submit_request(IORequest& req)	// Submit the request
  	{ to_serviceIORs.append(req); }
  int initiate_io(RID rid);		// Initiate the I/O for the request

					// Read/write a word from/to the 
					// PHYSICAL memory associated with 
					// the channel
					// Error is fatal
  void read_word_phys_memory(const Address ra, Word& dest);
  void write_word_phys_memory(const Address ra, Word src);
};

#endif
